// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt.plat;

import org.xwt.*;
import org.xwt.util.*;
import java.io.*;

/** common superclass for all platforms that use GCJ to compile a native binary */
public abstract class GCJ extends Platform {

    // static references to these classes ensure that the linker will include them
    private static Class c1 = gnu.java.locale.Calendar.class;
    private static Class c2 = java.util.GregorianCalendar.class;
    private static Class c3 = gnu.gcj.convert.Input_ASCII.class;
    private static Class c4 = gnu.gcj.convert.Input_UTF8.class;
    private static Class c5 = gnu.gcj.convert.Input_8859_1.class;
    private static Class c6 = gnu.java.locale.LocaleInformation.class;
    private static Class c7 = gnu.gcj.convert.Output_ASCII.class;
    private static Class c8 = gnu.java.locale.Calendar_en.class;
    private static Class c9 = gnu.java.locale.Calendar_en_US.class;

    protected org.xwt.Weak _getWeak(Object o) { return new Java2Weak(o); }

    private static class Java2Weak extends java.lang.ref.WeakReference implements org.xwt.Weak {
        public Java2Weak(Object o) { super(o); }
    }
        
    // FIXME
    protected ImageDecoder _decodeJPEG(InputStream is, String name) {
        try {
            return new JPEG(is);
        } catch (Exception e) {
            Log.log(this, "Exception while decoding JPEG image " + name);
            Log.log(this, e);
            return null;
        }
    }

    /** Converts an InputStream carrying a JPEG image into an ARGB int[] */
    private static class JPEG extends ImageDecoder {
        int[] data;
        byte[] buffer;
        int width, height;
        InputStream is;
        
        public final int[] getData() { return data; }
        public final int getWidth() { return width; }
        public final int getHeight() { return height; }
        private JPEG(InputStream is) { this.is = is; nativeDecompress(); buffer = null; }
        private native void nativeDecompress();
    }
    
    // FIXME: This could be optimized (a lot) by using a custom hashtable
    public final static class Retainer {
        private static Hash table = new Hash();
        private final static class Key {
            private Object o;
            public Key(Object o) { this.o = o; }
            public int hashCode() { return o == null ? 0 : o.hashCode(); }
            public boolean equals(Object o2) { return (o2 instanceof Key) && ((Key)o2).o == o; }
        }
        private final static class Entry {
            private int refCount;
            public Entry() { inc(); }
            public void inc() { refCount++; }
            public boolean dec() { return --refCount == 0; }
        }
        public static synchronized void retain(Object o) {
            Key k = new Key(o);
            Entry r = (Entry) table.get(k);
            if(r == null) table.put(k,new Entry());
            else r.inc();
        }
        public static synchronized void release(Object o) {
            Key k = new Key(o);
            Entry e = (Entry) table.get(k);
            if(e == null) throw new Error("Retainer::Release on unknown object");
            if(e.dec()) { table.remove(k); }
        }
    }
}
